from .typing import JSON
from .io_utils import write_json_to_stderr
from .async_io_utils import async_write_json_to_stderr


class REPORT_LEVEL:
    EXCEPTION = 50
    CRITICAL = 40
    ERROR = 30
    WARNING = 20
    INFO = 10
    DEBUG = 0


def _make_report_(
        original_data: JSON,
        log_level: int = REPORT_LEVEL.INFO,
        data: dict = None,
        message: str = None
) -> dict:

    report = {
        "logLevel": log_level
    }

    if message:
        report["message"] = message

    if data:
        report["data"] = data

    if not original_data:
        original_data = {"_meta": {"reporting": report}}

    else:
        original_data.setdefault("_meta", {}).setdefault("reporting", report)

    return original_data


def report(
        original_data: JSON = None,
        log_level: int = REPORT_LEVEL.INFO,
        data: JSON = None,
        message: str = None
):
    write_json_to_stderr(_make_report_(
        original_data,
        log_level,
        data,
        message
    ))

def eprint(message: str):
    report(message=message, log_level=REPORT_LEVEL.INFO)


def report_exception(
        original_data: JSON,
        exception: Exception,
        message: str = None
):
    write_json_to_stderr(_make_report_(
        original_data,
        log_level=REPORT_LEVEL.EXCEPTION,
        data={
            "exceptionDetails": str(exception)
        },
        message=message
    ))


async def async_report_exception(
        original_data: JSON or None,
        exception: Exception,
        message: str = None
):
    await async_write_json_to_stderr(_make_report_(
        original_data,
        log_level=REPORT_LEVEL.EXCEPTION,
        data={
            "exceptionDetails": str(exception)
        },
        message=message
    ))


__all__ = (
    "report", "REPORT_LEVEL", "report_exception", "async_report_exception",
    "eprint"
)



import sys
import builtins

from typing import Iterable, Tuple

from aioconsole import ainput, aprint, get_standard_streams

from .typing import JSON
from .exceptions import NotConnectedToPipe
from .json_utils import read_json, dump_json


async def async_read_stdin_lines(read_timeout: int = 0) -> Iterable[str]:
    """
    Read stdin line by line and return the tuple:

    (IS_STDIN_PIPE, IS_STDOUT_PIPE, LINE)

    :param read_timeout: read timeout of stdin. 0 = infinite
    :type read_timeout: int
    """

    if sys.stdin.isatty():
        raise NotConnectedToPipe(
            "Input data must be entered as a UNIX pipeline. For example: "
            "'cat info.json | tool-name'")

    is_stdout_pipe = not sys.stdout.isatty()

    # -------------------------------------------------------------------------
    # Read info by stdin or parameter
    # -------------------------------------------------------------------------
    while True:
        try:
            message = await ainput()

            yield is_stdout_pipe, message

        except EOFError:
            return


async def async_read_json_from_stdin() -> Tuple[bool, JSON]:
    # Read from the STDIN PIPE
    async for has_stdout_pipe, json_line in async_read_stdin_lines():
        try:
            yield False, read_json(json_line)

        except:
            yield True, json_line


async def async_write_json_to_stdout(
        data: JSON,
        force_flush: bool = True
):
    await async_write_to_stdout(dump_json(data), force_flush=force_flush)


async def async_write_json_to_stderr(data: JSON):
    await async_write_to_stderr(dump_json(data))


async def async_write_to_stdout(data: str, force_flush: bool = True):
    try:
        await aprint(data, flush=force_flush)
    except BrokenPipeError as e:
        print(
            f"[ERROR] when try to write data '{data}' in pipe: {e}",
            file=sys.stderr,
            flush=True
        )


async def async_write_to_stderr(data: str):
    if w := getattr(builtins, "cached_async_writer", None):
        writer = w
    else:
        _, writer = await get_standard_streams(use_stderr=True)
        builtins.cached_async_writer = w

    writer.write(f"{data}\n")
    await writer.drain()


__all__ = ("async_read_stdin_lines", "async_read_json_from_stdin",
           "async_write_json_to_stdout", "async_write_json_to_stderr",
           "async_write_to_stdout", "async_write_to_stderr")



try:
    import orjson as json

    JSON_LIBRARY = "orjson"

except ImportError:
    import json

    JSON_LIBRARY = "json"

from .typing import JSON

def dump_json(obj: dict or list) -> str:
    """Dumps json using most efficient JSON library available"""
    dump_text = json.dumps(obj)

    if JSON_LIBRARY == "orjson":
        dump_text = dump_text.decode()

    return dump_text

def read_json(data: str or bytes) -> JSON:
    """Reads json using most efficient JSON library available"""
    if JSON_LIBRARY != "orjson" and hasattr(data, "decode"):
        data = data.decode()

    return json.loads(data)


__all__ = ("dump_json", "read_json")



import sys
import select

from typing import Iterable, Tuple

from .typing import JSON
from .exceptions import NotConnectedToPipe
from .json_utils import read_json, dump_json


def read_stdin_lines(read_timeout: int = 0) -> Iterable[str]:
    """
    Read stdin line by line and return the tuple:

    (IS_STDIN_PIPE, IS_STDOUT_PIPE, LINE)

    :param read_timeout: read timeout of stdin. 0 = infinite
    :type read_timeout: int
    """

    if sys.stdin.isatty():
        raise NotConnectedToPipe(
            "Input data must be entered as a UNIX pipeline. For example: "
            "'cat info.json | tool-name'")

    is_stdout_pipe = not sys.stdout.isatty()

    # -------------------------------------------------------------------------
    # Read info by stdin or parameter
    # -------------------------------------------------------------------------
    if read_timeout:
        if sys.stdin in select.select([sys.stdin], [], [], read_timeout)[0]:
            for line in sys.stdin.readlines():
                yield is_stdout_pipe, line

        else:
            yield is_stdout_pipe, None

    else:

        while 1:

            line = sys.stdin.readline()

            if not line:
                return

            yield is_stdout_pipe, line


def read_json_from_stdin() -> Iterable[Tuple[bool, JSON]]:
    # Read from the STDIN PIPE
    for _, line in read_stdin_lines():

        try:
            # this var contains JSON data in our format
            yield False, read_json(line)

        except:
            yield True, line


def write_json_to_stdout(
    data: JSON,
    force_flush: bool = True
):
    write_to_stdout(dump_json(data), force_flush=force_flush)


def write_json_to_stderr(data: JSON):
    write_to_stderr(dump_json(data))


def write_to_stdout(data: str, force_flush: bool = True):
    try:
        sys.stdout.write(f"{data}\n")
    except BrokenPipeError as e:
        print(
            f"[ERROR] when try to write data '{data}' in pipe",
            file=sys.stderr,
            flush=True
        )

    if force_flush:
        try:
            sys.stdout.flush()
        except BrokenPipeError:
            ...


def write_to_stderr(data: str):
    sys.stderr.write(f"{data}\n")

    try:
        sys.stderr.flush()
    except BrokenPipeError:
        ...


__all__ = ("read_json_from_stdin", "write_json_to_stdout", "write_to_stdout",
           "write_to_stderr", "write_json_to_stderr",
           "read_stdin_lines")
